#pragma once

#include<dune/pdelab/common/quadraturerules.hh>
#include<dune/pdelab/finiteelement/localbasiscache.hh>
#include<dune/pdelab/localoperator/defaultimp.hh>
#include<dune/pdelab/localoperator/pattern.hh>
#include<dune/pdelab/localoperator/flags.hh>
#include<dune/pdelab/localoperator/variablefactories.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>

#include<dune/phasefield/localoperator/fem_base_chns.hh>


//Parameter class (for functions), FEM space and RangeField
template<typename Param, typename CHGFS, typename CHContainer, typename NSGFS, typename NSContainer>
class FEM_FS_CHNS_R_CahnHilliard :
  public Dune::PDELab::FullVolumePattern,
  public Dune::PDELab::LocalOperatorDefaultFlags,
  public Dune::PDELab::NumericalJacobianVolume<FEM_FS_CHNS_R_CahnHilliard<Param, CHGFS, CHContainer, NSGFS, NSContainer> >,
  public Dune::PDELab::NumericalJacobianApplyVolume<FEM_FS_CHNS_R_CahnHilliard<Param, CHGFS, CHContainer, NSGFS, NSContainer> >,
  public Dune::PDELab::NumericalJacobianBoundary<FEM_FS_CHNS_R_CahnHilliard<Param, CHGFS, CHContainer, NSGFS, NSContainer> >,
  public Dune::PDELab::NumericalJacobianApplyBoundary<FEM_FS_CHNS_R_CahnHilliard<Param, CHGFS, CHContainer, NSGFS, NSContainer> >
{
private:
  typedef typename CHContainer::field_type RF;
  typedef typename CHGFS::Traits::GridViewType::template Codim<0>::Entity   ENTITY;

  CHLocalBasisCache<CHGFS> chCache;
  VLocalBasisCache<NSGFS> vCache;

  typedef GFS_LocalViewer<CHGFS, CHContainer, ENTITY, RF> CHLView;
  mutable CHLView chOldLocal;

  typedef GFS_LocalViewer<NSGFS, NSContainer, ENTITY, RF> NSLView;
  mutable NSLView nsLocal;


  Param& param; // parameter functions

public:
  enum {doPatternVolume=true};
  enum {doAlphaVolume=true};
  enum { doAlphaBoundary = true };

  //constructor
  FEM_FS_CHNS_R_CahnHilliard(Param& param_, CHGFS& chGfs_, CHContainer& chVOld_, NSGFS& nsGfs_, NSContainer& nsV) :
    chOldLocal(chGfs_,chVOld_),
    nsLocal(nsGfs_,nsV),
    param(param_)
  {
  }


  template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
  void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
  {
    //Quadrature Rule
    const int dim = EG::Entity::dimension;
    const int order = 5; //TODO: Choose more educated
    auto rule = Dune::PDELab::quadratureRule(eg.geometry(), order);

    chOldLocal.DoEvaluation(eg);
    //chOldLocal.DoEvaluation(eg);
    nsLocal.DoSubEvaluation(eg,child(nsLocal.lfs,Dune::Indices::_0));

    for(const auto& ip : rule)
    {
      const auto S = eg.geometry().jacobianInverseTransposed(ip.position());
      auto factor = ip.weight()*eg.geometry().integrationElement(ip.position());

      CHVariables_U_PFOld_Extension<RF,dim,LFSV> ch(chCache, ip.position(), lfsv, x.base(),chOldLocal.lv, S);
      NSVariables<RF,dim,decltype(nsLocal.lfs)> ns(vCache, ip.position(), nsLocal.lfs, nsLocal.lv, S);

      RF pf_c = ch.pf;
      RF pf_f = ch.pf + 2*param.delta*(1-ch.pf);
      Dune::FieldVector<RF,dim> gradpf_f(0.0);
      gradpf_f.axpy((1-2*param.delta),ch.gradpf);

      RF pf_c_Old = ch.pfOld;
      RF pf_c_reg = pf_c+param.delta;
      RF pf_c_reg_Old = pf_c_Old + param.delta;

      for (size_t i=0; i<ch.pfspace.size(); i++)
      {
        RF Rspeed = param.dw.eval_q((ch.pfOld-0.05)/0.7)/0.7;
        RF R1 = -Rspeed/param.eps*
                  (param.reaction.eval(ch.u)+param.phys.curvatureAlpha*ch.xi)*ch.basis.phi[i]; //TODO Scale with sigma13
        RF J1 = param.mobility.eval(ch.pf)*param.eps * (ch.gradxi*ch.basis.gradphi[i][0]);

      // integrate: (pf-pfold)*phi_i/delta_t + gradpf * grad phi_i + ...
        r.accumulate(ch.pfspace,i,factor*(
          (ch.pf-ch.pfOld)*ch.basis.phi[i]/param.time.dt
          +J1
          //+(pf1*div_v + (gradpf1*v))*phi[i] TODO ADD THIS AGAIN WHEN BC ARE FIXED
          -R1
        ));

        r.accumulate(ch.xispace,i,factor*(
          ch.xi*ch.basis.phi[i]
          - 1/param.eps* param.dw.eval_d(ch.pf) *ch.basis.phi[i]
          - param.eps * (ch.gradpf*ch.basis.gradphi[i][0])
        ));


      // integrate (pf-pfold)*(u-uast)*phi_i/delta_t + pfold*(u-uold)*phi_i/delta_t + D*pf*gradu*grad phi_i = 0
        r.accumulate(ch.uspace,i,factor*(

          //(u-param.uAst)*phi[i]/param.dt //For Testing

          (pf_c_reg-pf_c_reg_Old)*(ch.u-param.phys.uAst)*ch.basis.phi[i]/param.time.dt
          + pf_c_reg_Old*(ch.u-ch.uOld)*ch.basis.phi[i]/param.time.dt
          + param.phys.D*pf_c_reg_Old*(ch.gradu*ch.basis.gradphi[i][0])
          + J1*(ch.u-param.phys.uAst)
          //+ (pf_c*(ch.u-param.phys.uAst)*ns.div_v + (ch.u-param.phys.uAst)*(ch.gradpf*ns.v) + pf_c*(ch.gradu*ns.v) )*ch.basis.phi[i]
          + ( pf_f*(ns.v*ch.gradu) )*ch.basis.phi[i]
          - pf_c_reg*param.reaction.eval_bulkreaction()*ch.basis.phi[i]//Bulk reaction
          ));
      }
    }
  }

  // boundary integral
  template<typename IG, typename LFSU, typename X, typename LFSV, typename R>
  void alpha_boundary (const IG& ig,
                       const LFSU& lfsu_s, const X& x_s, const LFSV& lfsv_s,
                       R& r_s) const
  {
    const int dim=2;
    // Define types
    using size_type = typename LFSV::Traits::SizeType;

    const auto& pfspace = child(lfsv_s,Dune::Indices::_0);

    // Reference to the inside cell
    const auto& cell_inside = ig.inside();

    // Get geometry
    auto geo = ig.geometry();
    auto ref_el_ig = referenceElement(geo);
    auto face_center_ig = ref_el_ig.position(0,0);

    // Get geometry of intersection in local coordinates of cell_inside
    auto geo_in_inside = ig.geometryInInside();

    auto geo_inside = cell_inside.geometry();

    // evaluate boundary condition type
    auto ref_el = referenceElement(geo_in_inside);
    auto local_face_center = ref_el.position(0,0);
    auto intersection = ig.intersection();

    Dune::FieldVector<RF, dim> xg = ig.geometry().global( face_center_ig);

    if(xg[0]>1e-7) return; //Only non-dirichlet at left brdy

    // loop over quadrature points and integrate normal flux
    auto intorder = 4;
    for (const auto& ip : quadratureRule(geo,intorder))
      {
        auto local = geo_in_inside.global(ip.position());
        // position of quadrature point in local coordinates of element
        const auto& S = geo_inside.jacobianInverseTransposed(local);

        auto n = ig.unitOuterNormal(ip.position());

        CHVariables<RF,dim,LFSV> ch(chCache, local, lfsv_s, x_s.base(), S);

        auto factor = ip.weight()*geo.integrationElement(ip.position());
        for (size_type i=0; i<ch.pfspace.size(); i++)
          r_s.accumulate(ch.xispace,i,0);//-(ch.gradpf*n)*ch.basis.phi[i]*factor);

      }
  }
};
