#pragma once

#include<dune/pdelab/common/quadraturerules.hh>
#include<dune/pdelab/finiteelement/localbasiscache.hh>
#include<dune/pdelab/localoperator/defaultimp.hh>
#include<dune/pdelab/localoperator/pattern.hh>
#include<dune/pdelab/localoperator/flags.hh>
#include<dune/pdelab/localoperator/variablefactories.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>

#include<dune/phasefield/localoperator/fem_base_chns.hh>


//Parameter class (for functions), FEM space and RangeField
template<typename Param, typename CHGFS, typename CHContainer, typename NSGFS, typename NSContainer>
class FEM_FS_R_Only :
  public Dune::PDELab::FullVolumePattern,
  public Dune::PDELab::LocalOperatorDefaultFlags,
  public Dune::PDELab::NumericalJacobianVolume<FEM_FS_R_Only<Param, CHGFS, CHContainer, NSGFS, NSContainer> >,
  public Dune::PDELab::NumericalJacobianApplyVolume<FEM_FS_R_Only<Param, CHGFS, CHContainer, NSGFS, NSContainer> >
{
private:
  typedef typename CHContainer::field_type RF;
  typedef typename CHGFS::Traits::GridViewType::template Codim<0>::Entity   ENTITY;

  CHLocalBasisCache<CHGFS> chCache;

  typedef GFS_LocalViewer<CHGFS, CHContainer, ENTITY, RF> CHLView;
  mutable CHLView chOldLocal;

  Param& param; // parameter functions

public:
  enum {doPatternVolume=true};
  enum {doAlphaVolume=true};

  //constructor
  FEM_FS_R_Only(Param& param_, CHGFS& chGfs_, CHContainer& chVOld_, NSGFS& nsGfs_, NSContainer& nsV) :
    chOldLocal(chGfs_,chVOld_),
    param(param_)
  {
  }


  template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
  void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
  {
    //Quadrature Rule
    const int dim = EG::Entity::dimension;
    const int order = 5; //TODO: Choose more educated
    auto rule = Dune::PDELab::quadratureRule(eg.geometry(), order);

    chOldLocal.DoEvaluation(eg);

    for(const auto& ip : rule)
    {
      const auto S = eg.geometry().jacobianInverseTransposed(ip.position());
      auto factor = ip.weight()*eg.geometry().integrationElement(ip.position());

      CHVariables_U_PFOld_Extension<RF,dim,LFSV> ch(chCache, ip.position(), lfsv, x.base(),chOldLocal.lv, S);

      for (size_t i=0; i<ch.pfspace.size(); i++)
      {
        RF Rspeed = param.dw.eval_q((ch.pfOld-0.05)/0.7)/0.7;
        RF R1 = -Rspeed/param.eps*
                  (param.reaction.eval(ch.uOld))*ch.basis.phi[i];

        r.accumulate(ch.pfspace,i,factor*(
          (ch.pf-ch.pfOld)*ch.basis.phi[i]/param.time.dt
          -R1
        ));

        r.accumulate(ch.xispace,i,factor*(
          (ch.xi-0)*ch.basis.phi[i]
        ));

        r.accumulate(ch.uspace,i,factor*(
          (ch.u-ch.uOld)*ch.basis.phi[i]
        ));
      }
    }
  }
};
