#pragma once

#include<dune/phasefield/localoperator/fem_base_chns.hh>


template<typename RF, const int dim, typename NSLFS>
class NSPorenetworkFlow
{
public:
  const typename NSLFS::template Child<0>::Type& vxspace;
  BasisFN<typename NSLFS::template Child<0>::Type> vbasis;

  RF vx;
  Dune::FieldVector<RF,dim> gradvx;

  template<typename Pos, typename vCache, typename LV, typename S>
  NSPorenetworkFlow(const vCache& vcache, const Pos& position, const NSLFS& nsLfs, const LV& lv, S& s) :
    vxspace(child(nsLfs,Dune::Indices::_0)),
    vbasis(vcache, position, vxspace, s),
    vx(0.0),
    gradvx(0.0)
  {
    for (size_t i=0; i<this->vxspace.size(); i++)
    {
      vx += lv[vxspace.localIndex(i)]*vbasis.phi[i];
      gradvx.axpy(lv[vxspace.localIndex(i)],vbasis.gradphi[i][0]);
    }
  }
};


template<typename NSGFS>
using Micro_VXLocalBasisCache =
  Dune::PDELab::LocalBasisCache<typename NSGFS::template Child<0>::Type::Traits::FiniteElementType::Traits::LocalBasisType>;
